from fpy.data.maybe import (
    Maybe,
    Just,
    Nothing,
    isJust,
    isNothing,
    fromJust,
    fromMaybe,
    maybe,
    mapMaybe,
)
from fpy.data.forgetful import forget, Under

import unittest


def add1(x):
    return x + 1


def even(n: int) -> Maybe[int]:
    if n % 2 == 0:
        return Just(n)
    return Nothing()


class TestMaybe(unittest.TestCase):
    def testJust(self):
        v = Just(3)
        self.assertTrue(isJust(v))
        self.assertFalse(isNothing(v))
        self.assertEqual(3, fromJust(v))
        self.assertEqual(3, fromMaybe(0, v))
        self.assertEqual(4, maybe(0, add1, v))

    def testNothing(self):
        v = Nothing()
        self.assertFalse(isJust(v))
        self.assertTrue(isNothing(v))
        with self.assertRaises(AssertionError):
            fromJust(v)
        self.assertEqual(3, fromMaybe(3, v))
        self.assertEqual(0, maybe(0, add1, v))

    def testList(self):
        lst = [1, 2, 3, 4, 5, 6, 7, 8]
        res = mapMaybe(even, lst)
        self.assertListEqual(res, [2, 4, 6, 8])

    def testBindJust(self):
        v = Just(4)
        vb = v >> even
        self.assertTrue(isJust(vb))
        self.assertEqual(4, fromJust(vb))

    def testBindJust1(self):
        v = Just(3)
        vb = v >> even
        self.assertTrue(isNothing(vb))

    def testBindNothing(self):
        n = Nothing()
        nb = n >> even
        self.assertTrue(isNothing(nb))
        self.assertIs(n, nb)

    def testNTransJust(self):
        v = Just(4)
        vforgot = v & forget
        self.assertIsInstance(vforgot, Under)
        self.assertEqual(vforgot.under(), 4)

    def testNTransNothing(self):
        v = Nothing()
        vforgot = v & forget
        self.assertIsInstance(vforgot, Under)
        self.assertIsNone(vforgot.under())



import inspect
import collections.abc as cabc

from fpy.composable.composable import Composable
from fpy.composable.transparent import Transparent

from typing import Callable

import sys
import traceback

class SignatureMismatchError(Exception):
    def __init__(self, e):
        self.e = e


class NotEnoughArgsError(Exception):
    def __init__(self, expect, got):
        self.expect = expect
        self.got = got


class func(Composable, Transparent):
    fn: Callable = None

    def __init__(self, f, *args, **kwargs):
        if isinstance(f, func):
            self.fn = f.fn
            self.args = (*f.args, *args)
            self.kwargs = {**f.kwargs, **kwargs}
            self.sig = f.sig
        elif isinstance(f, cabc.Callable):
            self.fn = f
            self.args = args
            self.kwargs = kwargs
            self.sig = inspect.signature(f)
        else:
            raise TypeError(f"{f} is not callable")

    def __underlying__(self):
        return self.fn

    def __repr__(self):
        return repr(self.fn)

    def __call__(self, *args, **kwargs):
        _args = (*self.args, *args)
        _kwargs = {**self.kwargs, **kwargs}
        try:
            self.sig.bind(*_args, **_kwargs)
            return self.fn(*_args, **_kwargs)
        except TypeError as e:
            # print(f"TypeError: {e}", file=sys.stderr)
            # tb = sys.exc_info()[2]
            # print(f"{tb.tb_frame.f_code.co_filename = }, {tb.tb_lineno = }")
            # traceback.print_tb(tb)
            try:
                self.sig.bind_partial(*_args, **_kwargs)
                return func(self, *args, **kwargs)
            except TypeError as e:
                raise SignatureMismatchError(e)
        except NotEnoughArgsError:
            return func(self, *args, **kwargs)

    def __compose__(self, other):
        if isinstance(other, func):
            return self.__class__(lambda *args, **kwargs: other(self(*args, **kwargs)))
        if isinstance(other, cabc.Callable):
            fn = func(other)
            return self.__class__(lambda *args, **kwargs: fn(self(*args, **kwargs)))
        raise TypeError(f"Cannot compose function with {type(other)}")



from __future__ import annotations

from fpy.control.monad import _Monad
from fpy.control.natural_transform import _NTrans
from abc import ABCMeta, abstractmethod, abstractclassmethod
import collections.abc as cabc
from typing import TypeVar, Generic, List, Callable
from dataclasses import dataclass

T = TypeVar("T")
S = TypeVar("S")
G = TypeVar("G")


class Maybe(_Monad[T], Generic[T]):
    @abstractmethod
    def __bool__(self):
        raise NotImplementedError

    def __bind__(self, b):
        raise NotImplementedError


@dataclass
class Just(Maybe[T], Generic[T]):
    v: T

    def __bool__(self):
        return True

    def __fmap__(self, f: Callable[[T], S]) -> Just[S]:
        return Just(f(self.v))

    def __ntrans__(self, t: _NTrans[Maybe, T, G, S]) -> G[S]:
        return t(self.v)

    def __bind__(self, b: Callable[[T], Maybe[S]]) -> Maybe[S]:
        return b(self.v)

    def __enter__(self):
        return self.v


class Nothing(Maybe[T], Generic[T]):
    def __bool__(self):
        return False

    def __bind__(self, _: any) -> Nothing[T]:
        return self

    def __ntrans__(self, t: _NTrans[Maybe, T, G, S]) -> G[S]:
        return t(None)

def isJust(m: Maybe[T]) -> bool:
    return isinstance(m, Just)


def isNothing(m: Maybe[T]) -> bool:
    return isinstance(m, Nothing)


def fromJust(m: Just[T]) -> T:
    assert isJust(m)
    return m.v


def fromMaybe(d: T, m: Maybe[T]) -> T:
    return d if not isJust(m) else fromJust(m)


def maybe(d: S, f: Callable[[T], S], m: Maybe[T]) -> S:
    return d if not isJust(m) else f(fromJust(m))


def mapMaybe(f: Callable[[T], Maybe[S]], l: cabc.Sequence[T]) -> List[S]:
    return [fromJust(v) for v in map(f, l) if isJust(v)]



from __future__ import annotations

from fpy.control.monad import _Monad
from fpy.composable.function import func, SignatureMismatchError
from fpy.composable.collections import get0, get1
from fpy.data.function import id_, const

import bytecode as bc

from dataclasses import dataclass
from typing import TypeVar, Generic, Callable, Tuple

import sys

T = TypeVar("T")
R = TypeVar("R")
S = TypeVar("S")

@dataclass
class State(_Monad[S], Generic[S, T]):
    comp : Callable[[S], Tuple[T, S]]

    @classmethod
    def ret(cls, v : T):
        def res(s : S) -> Tuple[T, S]:
            return v, s
        return cls(res)

    def __bind__(self, b : Callable[[T], State[S, R]]) -> State[S, R]:
        def res(s: S):
            _v, _s = runState(self, s)
            return runState(b(_v), _s)
        return State(res)

def runState(c : State[S, T], s : S) -> Tuple[T, S]:
    return c.comp(s)

def get() -> State[S, S]:
    return State(lambda s: (s, s))

def put(x: S) -> State[None, S]:
    return State(const((None, x)))

def modify(f : Callable[[S], S]) -> State[S, None]:
    return get() >> (lambda x: put(f(x)))

def gets(f : Callable[[S], T]) -> State[S, T]:
    return get() >> (lambda x: State.ret(f(x)))

def evalState(comp : State[S, T], s : S) -> T:
    return get0(runState(comp, s))

def execState(comp : State[S, T], s : S) -> T:
    return get1(runState(comp, s))



from fpy.composable.function import func

from typing import TypeVar, Callable, Any

T = TypeVar("T")
F = TypeVar("F")
A = TypeVar("A")
B = TypeVar("B")


def id_(x: T) -> T:
    return x


@func
def const(x: T, _: Any) -> T:
    return x


@func
def flip(f: Callable[[B, A], T], a: A, b: B) -> T:
    return f(b, a)


def fix(f: Callable) -> Callable:
    fn = func(f)
    return lambda *args: fn(fix(fn))(*args)


@func
def on(b: Callable[[B, B], T], u: Callable[[A], B], x: A, y: A) -> T:
    return b(u(x), u(y))

@func
def constN(n: int, x):
    if n == 1:
        return const(x)
    return const(constN(n - 1, x))

def uncurryN(n: int, f):
    def _res(*args):
        if len(args) != n:
            raise TypeError(f"Uncurried function {f} expected {n} arguments but {len(args)} was given")
        part = f
        for i in range(n):
            part = part(args[i])
        return part

    return _res



# poor man's quick lambda
# don't want to do the bytecode tricks lol

from fpy.composable.function import func


class IntermediateFunc(func):
    def __init__(self, fn):
        super().__init__(fn)

    def __eq__(self, a):
        return lambda x: self(x) == a

    def __ne__(self, a):
        return lambda x: self(x) != a

    def __add__(self, n):
        return lambda x: self(x) + n

    def __sub__(self, n):
        return lambda x: self(x) - n

    def __mul__(self, n):
        return lambda x: self(x) * n

    def __div__(self, n):
        return lambda x: self(x) / n

    def __radd__(self, n):
        return lambda x: n + self(x)

    def __rsub__(self, n):
        return lambda x: n - self(x)

    def __rmul__(self, n):
        return lambda x: n * self(x)

    def __rdiv__(self, n):
        return lambda x: n / self(x)

    def __getattr__(self, *args, **kwargs):
        return func(lambda x: self(x).__getattribute__(*args, **kwargs))

    def __getitem__(self, *args, **kwargs):
        return func(lambda x: self(x).__getitem__(*args, **kwargs))

    def __or__(self, other):
        return lambda x: self(x).__or__(other)

    def __hash__(self):
        return hash(id(self))


class Placeholder:
    def __eq__(self, a):
        return IntermediateFunc(lambda x: x == a)

    def __ne__(self, a):
        return IntermediateFunc(lambda x: x != a)

    def __add__(self, n):
        return IntermediateFunc(lambda x: x + n)

    def __sub__(self, n):
        return IntermediateFunc(lambda x: x - n)

    def __mul__(self, n):
        return IntermediateFunc(lambda x: x * n)

    def __div__(self, n):
        return IntermediateFunc(lambda x: x / n)

    def __radd__(self, n):
        return IntermediateFunc(lambda x: n + x)

    def __rsub__(self, n):
        return IntermediateFunc(lambda x: n - x)

    def __rmul__(self, n):
        return IntermediateFunc(lambda x: n * x)

    def __rdiv__(self, n):
        return IntermediateFunc(lambda x: n / x)

    def __getattr__(self, *args, **kwargs):
        return IntermediateFunc(lambda x: x.__getattribute__(*args, **kwargs))

    def __getitem__(self, *args, **kwargs):
        return IntermediateFunc(lambda x: x.__getitem__(*args, **kwargs))

    def __hash__(self):
        return hash(id(self))

    def __or__(self, other):
        return IntermediateFunc(lambda x: x.__or__(other))


__ = Placeholder()



from fpy.composable.function import func, NotEnoughArgsError
from fpy.data.forgetful import Under
from fpy.data.function import const

import collections.abc as cabc


def get_item(c, *args, **kwargs):
    return c.__getitem__(*args, **kwargs)


class Seq(func):
    def __init__(self, l):
        assert isinstance(l, cabc.Sequence), f"{l} is not a sequence"
        self._l = l
        super().__init__(get_item, l)

    def __getattr__(self, *args, **kwargs):
        return self._l.__getattribute__(*args, **kwargs)

    def __getitem__(self, *args, **kwargs):
        return self._l.__getitem__(*args, **kwargs)

    def __str__(self):
        return f"#Seq{str(self._l)}"

    def __repr__(self):
        return f"#Seq{repr(self._l)}"

    def split(self, pred):
        a = []
        b = []
        for v in self._l:
            if pred(v):
                a.append(v)
            else:
                b.append(v)
        return a, b

    def filter(self, pred):
        return get0(self.split(pred))


class Map(func):
    def __init__(self, m):
        assert isinstance(m, cabc.Mapping), f"{m} is not a map"
        self._m = m
        super().__init__(get_item, m)

    def __getattr__(self, *args, **kwargs):
        return self._m.__getattribute__(*args, **kwargs)

    def __getitem__(self, *args, **kwargs):
        return self._m.__getitem__(*args, **kwargs)

    def __str__(self):
        return f"#Map{str(self._m)}"

    def __repr__(self):
        return f"#Map{repr(self._m)}"

    def split(self, pred):
        a = {}
        b = {}
        for k, v in self._m.items():
            if pred(k):
                a[k] = v
            else:
                b[k] = v
        return a, b

    def filter(self, pred):
        return get0(self.split(pred))


@func
def transN(n, fn, it):
    return type(it)((fn(v) if _i == n else v for _i, v in enumerate(it)))


trans0 = transN(0)
trans1 = transN(1)


@func
def setN(n, v, it):
    return transN(n, const(v), it)


set0 = setN(0)
set1 = setN(1)


@func
def getN(n, x):
    return x[n]


get0 = getN(0)
get1 = getN(1)

def not_(pred):
    return func(lambda x: not pred(x))

def of_(*it):
    return func(lambda x: x in it)


def is_(ty):
    return func(lambda x: isinstance(x, ty))


def and_(a, b):
    @func
    def __and(*args, **kwargs):
        _a = a(*args, **kwargs)
        if _a:
            return b(*args, **kwargs)
        return _a

    return __and


def or_(a, b):
    @func
    def __or(*args, **kwargs):
        _a = a(*args, **kwargs)
        if not _a:
            return b(*args, **kwargs)
        return _a

    return __or


@func
def to(dst, src):
    return dst(src)


def apply(fn):
    @func
    def __apply(_a=None, _k=None):
        a = _a if _a is not None else ()
        k = _k if _k is not None else {}
        return fn(*a, **k)

    return __apply


@func
def mapN(n, fn, *lsts):
    if len(lsts) < n:
        raise NotEnoughArgsError(n, len(lsts))
    ap = apply(fn)
    return [ap(row) for row in zip(*lsts)]


mp1 = mapN(1)
mp2 = mapN(2)


@func
def eqN(n, it, x):
    return x == it[n]


eq0 = eqN(0)
eq1 = eqN(1)

fwd_ = Under.ret

def seq2map(*keys):
    @func
    def __seq2map(*vals):
        if len(vals) < len(keys):
            raise NotEnoughArgsError(len(keys), len(vals))
        return dict(zip(keys, vals))

    return __seq2map



from fpy.experimental.case import case

import unittest

class TestCase(unittest.TestCase):
    def testConstPat(self):
        @case
        def test(x):
            case(x)
            { 1: 1,
              2: 2 }

        self.assertEqual(1, test(1))
        self.assertEqual(2, test(2))



from fpy.data.state import execState
from fpy.data.state import State, runState, put, get, modify, gets, evalState
from fpy.data.function import const
from fpy.composable.function import func
from fpy.composable.collections import get1

from fpy.experimental.do import do
from fpy.experimental.case import case

import unittest

class TestState(unittest.TestCase):
    def testRet(self):
        r = runState(State.ret('X'), 1)
        self.assertEqual(('X', 1), r)
    
    def testGet(self):
        r = runState(get(), 1)
        self.assertEqual((1, 1), r)
    
    def testPut(self):
        r = runState(put(5), 1)
        self.assertEqual((None, 5), r)

    def testModify(self):
        r = runState(modify(lambda x: x + 1), 1)
        self.assertEqual((None, 2), r)

    def testPlay(self):
        # Example from: https://wiki.haskell.org/State_Monad
        # GameState = (Bool, Int)
        def play(chars):
            if chars == []:
                return gets(get1) >> State.ret

            @func
            def case(c, st):
                o, s = st
                if c == 'a' and o:
                    return put((o, s + 1))
                elif c == 'b' and o:
                    return put((o, s - 1))
                elif c == 'c':
                    return put((not o, s))
                else:
                    return put((o, s))

            x, *xs = chars

            return get() >> case(x) >> const(play(xs))

        r = evalState(play("abcaaacbbcabbab"), (False, 0))
        self.assertEqual(2, r)

    def testDo(self):
        @do(State)
        def test():
            s <- get()
            put(s + 1)
            modify(lambda x: x * 2)

        r = execState(test(), 1)
        self.assertEqual(4, r)

    def testPlayDo(self):
        # Example from: https://wiki.haskell.org/State_Monad
        # GameState = (Bool, Int)
        @do(State)
        def play(chars):
            (o, s) <- get()
            if chars == []:
                return s

            @func
            def case(c, o, s):
                if c == 'a' and o:
                    return put((o, s + 1))
                elif c == 'b' and o:
                    return put((o, s - 1))
                elif c == 'c':
                    return put((not o, s))
                else:
                    return put((o, s))

            x, *xs = chars

            case(x, o, s)
            play(xs)

        r = evalState(play("abcaaacbbcabbab"), (False, 0))
        self.assertEqual(2, r)

    def testPlayDoPat(self):
        # Example from: https://wiki.haskell.org/State_Monad
        # GameState = (Bool, Int)
        @do(State)
        @case
        def play(chars):
            (o, s) <- get()
            if chars == []:
                return s

            x, *xs = chars

            case(x, o, s)({
                ('a', True, _s) : put((o, _s + 1)),
                ('b', True, _s) : put((o, _s - 1)),
                ('c', _o, _s)   : put((not _o, _s)),
                _               : put((o, s))
                })
            play(xs)

        r = evalState(play("abcaaacbbcabbab"), (False, 0))
        self.assertEqual(2, r)



from fpy.data.maybe import (
    Maybe,
    Just,
    Nothing,
    isJust,
    isNothing,
    fromJust,
)
from fpy.control.monad import do

import unittest


def foo(x, y):
    return Just(x + y)


class TestDo(unittest.TestCase):
    def testSimpleJust(self):
        @do
        def test():
            with Just(1) as x:
                return Just(x)

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 1)

    def testSimpleNothing(self):
        @do
        def test():
            with Just(1) as x:
                return Nothing()

        res = test()
        self.assertTrue(isNothing(res))

    def testLocal(self):
        @do
        def test():
            x = Just(1)
            with x as y:
                return Just(y)

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 1)

    def testLocalNested(self):
        @do
        def test():
            x = 1
            with Just(2) as y:
                with Just(x + y) as z:
                    return Just(z)

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 3)

    def testNested(self):
        @do
        def test():
            with Just(1) as  x, Just(2) as y, foo(x, y) as z:
                return Just(z + z)

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 6)

    def testTuple(self):
        @do
        def test():
            with Just((1, 2)) as (a, b):
                return Just(a + b)

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 3)

    def testComplex(self):
        @do
        def test():
            with Just((1, 2)) as (a, b), Just(a + b) as c, Just(c * b) as d:
                return Just(d)

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 6)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import TypeVar, Generic, Callable


F = TypeVar("F")
G = TypeVar("G")
A = TypeVar("A")
B = TypeVar("B")


class _NTrans(Generic[F, A, G, B], metaclass=ABCMeta):
    @abstractmethod
    def __trans__(self, v: A) -> G[B]:
        raise NotImplementedError

    def __call__(self, v: A) -> G[B]:
        return self.__trans__(v)


@dataclass
class NTrans(_NTrans):
    f: Callable[[A], G[B]]

    def __trans__(self, v: F[A]) -> G[B]:
        assert hasattr(v, "__ntrans__")
        return v.__ntrans__(self.f)



from abc import ABCMeta, abstractmethod, abstractclassmethod


class Transparent(metaclass=ABCMeta):
    @abstractmethod
    def __underlying__(self):
        raise NotImplementedError

    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            return getattr(self.__underlying__(), name)



from fpy.data.either import Either, Left, Right
from fpy.data.forgetful import forget
from fpy.data.function import const, flip
from fpy.utils.placeholder import __
from fpy.composable.transparent import Transparent
from fpy.composable.function import func
from fpy.composable.collections import (
    apply,
    transN,
    trans0,
    trans1,
    or_,
    and_,
    set0,
    set1,
    get0,
    get1,
)

import string
from typing import TypeVar, List, Tuple, Callable, Generic, Sequence, Union, Optional
from dataclasses import dataclass
import collections.abc as cabc

S = TypeVar("S")
T = TypeVar("T")


def isspace(c):
    return c in string.whitespace


def s2c(s: str) -> List[str]:
    return list(s)


def c2s(cs: List[str]) -> str:
    return "".join(cs)


@dataclass
class parser(Transparent, Generic[S, T]):
    """
    parser :: [S] -> Either [S] ([T] * [S])
    """

    fn: Union[Callable[[Sequence[S]], Either[Sequence[S], Tuple[T, Sequence[S]]]], Callable[[Sequence[S]], Optional[Tuple[T, Sequence[S]]]]] 

    def __underlying__(self):
        return self.fn

    def __call__(self, s: Sequence[S]):
        if not isinstance(s, cabc.Sequence):
            raise TypeError("Cannot parse none sequence")
        if not s:
            return Left(s)
        res = self.fn(s)
        if res is None:
            return Left(s)
        assert isinstance(res, (tuple, Either)), f"{res} type is {type(res)}"
        return Right(res) if isinstance(res, tuple) else res

    def timeN(self, n):
        if n <= 0:
            return parser(const(None))

        p = self
        for n in range(n - 1):
            p = p + self

        return p

    def __mul__(self, n):
        return self.timeN(n)

    def __rmul__(self, n):
        return self.timeN(n)

    def concat(self, nxt):
        @parser
        def __concat(s):
            return self(s) >> apply(lambda a, rest: nxt(rest) | trans0(a + __))

        return __concat

    def __add__(self, nxt):
        return self.concat(nxt)

    def choice(self, other):
        @parser
        def __choice(s):
            return or_(self, other)(s) or None

        return __choice

    def __or__(self, other):
        return self.choice(other)

    def parseR(self, rightP):
        @parser
        def __parseR(s):
            return self(s) >> apply(lambda leftR, rest: rightP(rest))

        return __parseR

    def parseL(self, rightP):
        @parser
        def __parseL(s):
            return self(s) >> apply(lambda leftR, rest: rightP(rest) | set0(leftR))

        return __parseL

    def __rshift__(self, other):
        return self.parseR(other)

    def __lshift__(self, other):
        return self.parseL(other)


def one(pred):
    @parser
    def res(s):
        if pred(s[0]):
            return [s[0]], s[1:]
        return None

    return res


def neg(pred):
    @parser
    def res(s):
        if not pred(s[0]):
            return [s[0]], s[1:]
        return None

    return res


just_nothing = parser(lambda s: ([], s))
pmaybe = __ | just_nothing


def many1(p):
    @parser
    def __many1(s):
        _res = []
        while s:
            _part = p(s)
            if not _part:
                break
            part, s = (_part & forget).under()
            _res += part
        if not _res:
            return None
        return _res, s

    return __many1


many = lambda p: pmaybe(many1(p))


def ptrans(p, trans):
    return parser(lambda s: p(s) | trans)


def peek(p):
    @parser
    def __peek(s):
        return p(s) | set1(s)

    return __peek


discard = set0([])
skip = flip(ptrans, discard)


def pseq(s):
    if not s:
        return just_nothing
    p = just_nothing
    for e in s:
        p = p + one(__ == e)
    return p

def inv(p):
    @parser
    def __inv(s):
        _r = p(s)
        if _r:
            return None
        return [], s
    return __inv



import bytecode as bc

import pprint

from fpy.parsec.parsec import parser, ptrans, one, many
from fpy.composable.collections import (
    of_,
    is_,
    and_,
    or_,
    trans0,
    fwd_,
    get0,
    mp1,
    apply,
)
from fpy.data.function import id_, const
from fpy.data.forgetful import forget
from fpy.data.maybe import Just, isJust, fromJust
from fpy.data.either import Left, Right, rights, either
from fpy.utils.placeholder import __
from fpy.composable.function import func
from fpy.debug.debug import trace

from dataclasses import dataclass

import dis
import sys

pp = pprint.PrettyPrinter(indent=4)

@dataclass
class Then:
    lineno: int


@dataclass
class Arrow:
    lineno: int


@dataclass
class TupleArrow:
    lineno: int
    ntpl: int


@dataclass
class Ret:
    lineno: int


isInstr = is_(bc.Instr)
isArrow = is_(Arrow)
dash = and_(isInstr, __.name == "UNARY_NEGATIVE")
arrowHead = and_(isInstr, and_(__.name == "COMPARE_OP", __.arg == bc.Compare.LT))
popTop = and_(isInstr, __.name == "POP_TOP")
none = and_(isInstr, and_(__.name == "LOAD_CONST", __.arg == None))
mkTpl = and_(isInstr, __.name == "BUILD_TUPLE")
ret = and_(isInstr, __.name == "RETURN_VALUE")
load = and_(
    isInstr,
    __.name
    ^ of_(
        "LOAD_GLOBAL",
        "LOAD_NAME",
        "LOAD_FAST",
        "LOAD_CLOSURE",
        "LOAD_DEREF",
        "LOAD_CLASSDEREF",
    ),
)
store = and_(
    isInstr,
    __.name
    ^ of_(
        "STORE_FAST",
        "STORE_NAME",
        "STORE_DEREF",
    ),
)
parseArrow = ptrans(
    one(dash) << one(arrowHead) << one(popTop),
    trans0(trans0(__.lineno ^ Arrow)),
)
parseTplArrow = ptrans(
    one(dash) >> one(arrowHead) >> one(mkTpl) << one(popTop),
    trans0(trans0(lambda x: TupleArrow(x.lineno, x.arg))),
)
parseThen = ptrans(
        one(popTop),
        trans0(trans0(__.lineno ^ Then)),
)
parseNoneRet = many(
    ptrans(
        one(popTop) << one(none) << one(ret),
        trans0(trans0(__.lineno ^ Ret)),
    )
    | one(const(True))
)

parseDo = many(parseTplArrow | parseArrow | parseThen | one(const(True)))
isFast = and_(isInstr, __.name == "LOAD_FAST")
fastToCell = ptrans(
    one(isFast),
    trans0(trans0(__.arg ^ bc.CellVar ^ func(bc.Instr, "LOAD_DEREF"))),
)
isSFast = and_(isInstr, __.name == "STORE_FAST")
storeFastToCell = ptrans(
    one(isSFast),
    trans0(trans0(__.arg ^ bc.CellVar ^ func(bc.Instr, "STORE_DEREF"))),
)
transFast = many(fastToCell | storeFastToCell | one(const(True)))


@dataclass
class ArrowInst:
    comp: list
    argnames: list
    nxt: list


def transformRet(insts):
    res = []
    for inst in insts:
        if isinstance(inst, Ret):
            res.append(bc.Instr("RETURN_VALUE", lineno=inst.lineno))
            continue
        if ret(inst):
            res.append(bc.Instr("LOAD_DEREF", bc.CellVar("!ret"), lineno=inst.lineno))
            if sys.version_info.major == 3 and sys.version_info.minor >= 11:
                res.append(bc.Instr("PUSH_NULL"))
                res.append(bc.Instr("SWAP", 3, lineno=inst.lineno))
                res.append(bc.Instr("PRECALL", 1))
                res.append(bc.Instr("CALL", 1))
            else:
                res.append(bc.Instr("ROT_TWO", lineno=inst.lineno))
                res.append(bc.Instr("CALL_FUNCTION", 1, lineno=inst.lineno))
            # res.append(bc.Instr("DUP_TOP", lineno=inst.lineno))
            # res.append(bc.Instr("PRINT_EXPR", lineno=inst.lineno))
            res.append(bc.Instr("RETURN_VALUE", lineno=inst.lineno))
            continue
        res.append(inst)
    return res


def partitionInst(insts, n):
    if not insts:
        return [], []
    if n == 0:
        return [], insts
    head = insts[-1]
    # print(f"{head = }")
    # print(f"{n = }")
    pre, post = head.pre_and_post_stack_effect()
    # print(f"{pre = }")
    # print(f"{post= }")
    if pre >= 0:
        if pre == n:
            return [head], insts[:-1]
        if pre < n:
            nxt, rst = partitionInst(insts[:-1], n - pre)
            return nxt + [head], rst
    pre = abs(pre)
    nxt, rst = partitionInst(insts[:-1], pre)
    if post == n:
        return nxt + [head], rst
    if post < n:
        head = nxt + [head]
        nxt, rst = partitionInst(rst, n - post)
        return nxt + head, rst


def transformDo(insts):
    # pp.pprint(insts)
    res = []
    while insts:
        inst = insts.pop()
        if not isinstance(inst, (Then, Arrow, TupleArrow)):
            res.insert(0, inst)
            continue
        if isinstance(inst, Then):
            # a >> b := a >>= \_ -> b
            comp, insts = partitionInst(insts, 1)
            res = [ArrowInst(comp, [f"!_{inst.lineno}"], res)]
            continue
        if isinstance(inst, Arrow):
            comp, insts = partitionInst(insts, 1)
            bindName = insts.pop()
            assert load(bindName) or (
                isInstr(bindName) and bindName.name == "BUILD_TUPLE"
            ), "it has to be tuple of symbols on the left of <-"
            if load(bindName):
                name = bindName.arg
                if sys.version_info.major == 3 and sys.version_info.minor >= 11:
                    if isinstance(name, tuple):
                        name = name[1]
                res = [ArrowInst(comp, [name], res)]
                continue
            if isInstr(bindName) and bindName.name == "BUILD_TUPLE":
                narg = bindName.arg
                names = []
                for _ in range(narg):
                    name = insts.pop()
                    assert load(name), "it has to be tuple of symbols on the left of <-"
                    if sys.version_info.major == 3 and sys.version_info.minor >= 11 and isinstance(name.arg, tuple):
                        names.insert(0, name.arg[1]) 
                    else:
                        names.insert(0, name.arg)
                res = [ArrowInst(comp, names, res)]
                continue
        elif isinstance(inst, TupleArrow):
            comp, insts = partitionInst(insts, 1)
            names = []
            for _ in range(inst.ntpl):
                name = insts.pop()
                assert load(name), "it has to be tuple of symbols on the left of <-"
                if sys.version_info.major == 3 and sys.version_info.minor >= 11 and isinstance(name.arg, tuple):
                    names.insert(0, name.arg[1]) 
                else:
                    names.insert(0, name.arg)
            res = [ArrowInst(comp, names, res)]
            continue

    return res


def generateLocal(insts, argnames, fv, cv, name, filename):
    resBc = []
    bcobj = bc.Bytecode(resBc)
    bcobj.freevars = fv
    bcobj.cellvars = cv
    bcobj.name = name
    bcobj.filename = filename
    bcobj.docstring = "generated local function for binding in do notation"
    bcobj.argcount = 1
    bcobj.argname.extend(argnames)
    co = bcobj.to_code()
    return co


def build_do_func(name, arg, body, free):
    inner = doDeco(body, name, arg, free)
    return inner


def markArg(args):
    def res(x):
        if load(x):
            if sys.version_info.major == 3 and sys.version_info.minor >= 11 and isinstance(x.arg, tuple):
                if x.arg[1] in args:
                    return Right(bc.Instr("LOAD_CONST", x.arg[1]))
                else:
                    return Left(x)
            else:
                return Right(x) if load(x) and x.arg in args else Left(x)
        else:
            return Left(x)
    return res

toArg = __.arg ^ bc.CellVar ^ func(bc.Instr, "LOAD_DEREF")
# toArg = __.arg ^ func(bc.Instr, "LOAD_FAST")
isCell = and_(
    isInstr,
    and_(
        __.name ^ of_("LOAD_DEREF", "STORE_DEREF", "LOAD_CLOSURE"),
        __.arg ^ is_(bc.CellVar),
    ),
)
transFree = lambda free: (
    lambda x: bc.Instr("LOAD_DEREF", bc.FreeVar(x.arg))
    if and_(isInstr, and_(__.name == "LOAD_GLOBAL", __.arg ^ of_(*free)))(x)
    else x
)
markFree = lambda free: (
    lambda x: Right(x) if isCell(x) and x.arg.name in free else Left(x)
)
toFree = lambda x: bc.Instr(x.name, bc.FreeVar(x.arg.name))


@func
def doArrow(name, cells, free, arrow):
    if not isinstance(arrow, ArrowInst):
        return [arrow]
    # print(f"doing arrow: {arrow}")
    bind_fn_name = f"{name}.__do_bind_{'_'.join(arrow.argnames)}__"
    # print(f"{bind_fn_name = }")
    arrfn = build_do_func(bind_fn_name, arrow.argnames, arrow.nxt, [*cells, *free])
    # print("=" * 20)
    # dis.dis(arrfn)
    # dis.show_code(arrfn)
    # print("=" * 20)
    # print(f"{bind_fn_name = }")
    rawcomp = arrow.comp
    # print(f"{rawcomp = }")
    transcomp = (
        mp1(markArg([arrow.argnames, *free, *cells]))
        ^ mp1(either(id_, toArg))
        ^ mp1(markFree(free))
        ^ mp1(either(id_, toFree))
        ^ transFree(free)
    )(rawcomp)
    # print(f"{transcomp = }")
    res = [
        *transcomp,
        * ([bc.Instr("LOAD_METHOD", "__bind__", lineno=arrow.comp[-1].lineno)]
           if sys.version_info.major == 3 and sys.version_info.minor >= 11 else
           [bc.Instr("LOAD_ATTR", "__bind__", lineno=arrow.comp[-1].lineno)])
    ]
    ncells = len(cells)
    for cell in cells:
        res.append(bc.Instr("LOAD_CLOSURE", bc.CellVar(cell)))
    for f in free:
        if f not in cells:
            res.append(bc.Instr("LOAD_CLOSURE", bc.FreeVar(f)))
            ncells += 1
    if cells:
        res.append(bc.Instr("BUILD_TUPLE", ncells))
    res.append(bc.Instr("LOAD_CONST", arrfn))
    if not (sys.version_info.major == 3 and sys.version_info.minor >= 11):
        res.append(
            bc.Instr(
                "LOAD_CONST",
                bind_fn_name,
                lineno=arrow.comp[-1].lineno,
            )
        )
    res.append(bc.Instr("MAKE_FUNCTION", 0x08, lineno=arrow.comp[-1].lineno))
    if len(arrow.argnames) > 1:
        res.append(bc.Instr("LOAD_CONST", apply))
        if sys.version_info.major == 3 and sys.version_info.minor >= 11:
            res.append(bc.Instr("PUSH_NULL"))
            res.append(bc.Instr("SWAP", 3))
            res.append(bc.Instr("PRECALL", 1))
            res.append(bc.Instr("CALL", 1))
        else:
            res.append(bc.Instr("ROT_TWO"))
            res.append(bc.Instr("CALL_FUNCTION", 1))
    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
        res.append(bc.Instr("PRECALL", 1))
        res.append(bc.Instr("CALL", 1))
    else:
        res.append(bc.Instr("CALL_FUNCTION", 1, lineno=arrow.comp[-1].lineno))
    res.append(bc.Instr("RETURN_VALUE", lineno=arrow.comp[-1].lineno))
    return res

def makeCells(insts):
    cells = set()
    for inst in insts:
        if not isInstr(inst):
            continue
        if inst.name != 'MAKE_CELL':
            continue
        if inst.arg.name not in cells:
            cells.add(inst.arg.name)
    return cells

def doDeco(b, name, args, free):
    # print(f"{free = }")
    # print("RAW BC: ", b)
    res = (
        parseDo(b) >> (get0 ^ transFast) & forget
        | get0
        ^ mp1(markFree(free))
        ^ mp1(either(id_, toFree))
        ^ mp1(markArg(args))
        ^ mp1(either(id_, toArg))
        ^ mp1(transFree(free))
        # ^ trace("before trans do = ")
        ^ transformDo
    )
    # print("Trans BC: ", res.under())
    getCellName = lambda x: Right(x) if isCell(x) else Left(x)
    cells = res | mp1(getCellName) | rights | mp1(__.arg.name) | __ + list(args) | set
    # print(f"{cells = }")
    res = (
        res
        # | trace("res = ")
        | mp1(doArrow(name, cells.under(), free))
        # | trace("didArrow = ")
        | func(sum, start=[]) ^ bc.Bytecode
    )
    # print(f"{res.under() = }")
    resbc = res.under()

    madeCells = makeCells(resbc)

    # pp.pprint(resbc)
    resbc.cellvars.extend(cells.under())
    resbc.cellvars.extend(free)
    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
        for cell in resbc.cellvars:
            if cell not in madeCells:
                resbc.insert(0, bc.Instr("MAKE_CELL", bc.CellVar(cell)))
    resbc.freevars.extend(free)
    if len(resbc.freevars) > 0:
        resbc.insert(0, bc.Instr("COPY_FREE_VARS", len(resbc.freevars)))
    resbc.argcount = len(args)
    resbc.argnames.extend(args)
    resbc.name = name
    resbc.filename = name
    resbc.flags = resbc.flags | 16
    resbc.update_flags()
    co = resbc.to_code()
    # print("======================")
    # print(name)
    # print("======================")
    # dis.dis(co)
    # dis.show_code(co)
    return co


def do(m):
    def res(fn):
        ret = m
        # add `!ret` to the front of function
        retbc = [
            bc.Instr("LOAD_CONST", m),
            bc.Instr("LOAD_ATTR", "ret"),
            bc.Instr("STORE_DEREF", bc.CellVar("!ret")),
        ]
        rawbc = bc.Bytecode.from_code(fn.__code__)
        # print(f"{rawbc = }")
        args = fn.__code__.co_varnames[: fn.__code__.co_argcount]
        b = parseNoneRet(retbc + rawbc) & forget | get0 | transformRet
        co = doDeco(b.under(), fn.__name__, args, rawbc.freevars)
        fn.__code__ = co
        return fn

    return res



from __future__ import annotations

from abc import ABCMeta, abstractmethod, abstractclassmethod
from fpy.control.functor import _Functor

from dataclasses import dataclass
from typing import TypeVar, Generic, Callable

T = TypeVar("T")
S = TypeVar("S")


class _Applicative(_Functor[T], Generic[T]):
    @classmethod
    def pure(cls, val):
        return cls(val)

    @classmethod
    def liftA2(cls, f, a, b):
        return cls.fmap(cls.fmap(f, a), b)


@dataclass
class Applicative(_Applicative[T], Generic[T]):
    val: T

    def __fmap__(self, f: Callable[[T], S]) -> _Applicative[S]:
        return type(self)(f(self.val))



#!/usr/bin/env python3

VERSION = "0.0.18"

DESCRIPTION = "Python module for composing computations"
CLASSIFIERS = [
    "Intended Audience :: Developers",
    "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3",
    "Topic :: Software Development :: Libraries :: Python Modules",
]


def main():
    try:
        from setuptools import setup
    except ImportError:
        from distutils.core import setup

    with open("README.md") as fin:
        desc = fin.read().strip()

    options = {
        "name": "fppy",
        "version": VERSION,
        "license": "GPLv3",
        "description": DESCRIPTION,
        "long_description": desc,
        "long_description_content_type": "text/markdown",
        "url": "https://github.com/Z-Shang/fpy",
        "author": "zshang",
        "author_email": "z@gilgamesh.me",
        "classifiers": CLASSIFIERS,
        "packages": [
            "fpy",
            "fpy.composable",
            "fpy.control",
            "fpy.data",
            "fpy.debug",
            "fpy.experimental",
            "fpy.parsec",
            "fpy.utils",
            "fpy.tests",
        ],
        "install_requires": ["bytecode"],
    }
    setup(**options)


if __name__ == "__main__":
    main()
